------------------------------------------------------------------------------------
-- Keyer Project
-- Jim Brady   Jan 2011
-- The license is FreeBSD, see the file "copying.txt"

-- This is the top level file for the morse code keyer.
-- Speed is settable over RS232. It has has a speaker output tone
-- and an output for the CW transmitter. Uses the Starter Kit south
-- button for reset. 
--
-- Development tools:  Xilinx ISE WebPack 12.4
-- Target              Xilinx Spartan 3A Starter Kit Board
-- FPGA                Xilinx XC3S700A-4FG484
--
-- To set speed, use a RS232 terminal set to 115.2K, N, 8, 1
-- Connect it to the DB9 "DCE" (female) connector on the board.
-- Press '<' or '>' key to change speed by 1 wpm each time. 
-- After changing wpm, it sends it to terminal to confirm
------------------------------------------------------------------------------------

library IEEE;
   use IEEE.std_logic_1164.all;
   use IEEE.numeric_std.all;
   
   -- Include my components
   use work.my_components.all;


entity keyer is
   port (clk         : in  std_logic;  -- 50MHz clock
         reset       : in  std_logic;  -- South button
         dot_key     : in  std_logic;  -- Input from paddle
         dash_key    : in  std_logic;  -- Input from paddle
         uart_rx     : in  std_logic;  -- RS232 line in
		   uart_tx     : out std_logic;  -- RS232 line out
         key_out     : out std_logic;  -- Output to transmitter
         led0        : out std_logic;  -- LED on board
         spkr        : out std_logic); -- Output to speaker
end keyer;


architecture rtl of keyer is

          
   -- This lookup table is used to convert the WPM setting into
   -- limit values for the the dot timer and dash timer.  It is      
   -- based on a 50MHz FPGA clock.  See Readme.txt for info on
   -- Morse code timing
   type wpm_array is array(0 to 63) of integer;
   constant wpm_lookup: wpm_array :=  
       (12000000, 12000000, 12000000, 12000000, 12000000,
        12000000, 10000000,  8571429,  7500000,  6666667,
         6000000,  5454545,  5000000,  4615385,  4285714,
         4000000,  3750000,  3529412,  3333333,  3157895,
         3000000,  2857143,  2727273,  2608696,  2500000,
         2400000,  2307692,  2222222,  2142857,  2068966,
         2000000,  1935484,  1875000,  1818182,  1764706,
         1714286,  1666667,  1621622,  1578947,  1538462,
         1500000,  1463415,  1428571,  1395349,  1363636,
         1333333,  1304348,  1276596,  1250000,  1224490,
         1200000,  1176471,  1153846,  1132075,  1111111,
         1090909,  1071429,  1052632,  1034483,  1016949,
         1000000,   983607,   967742,   952381);

  	
	type state_type is (IDLE, DOT, DASH, SPACE_AFTER_DOT, SPACE_AFTER_DASH);
   signal state, next_state: state_type;
   
   signal wpm            : integer range 0 to 63;
	signal dot_time       : integer;
   signal dash_time      : integer;
	signal dot_mem        : std_logic;
   signal dash_mem       : std_logic;
   signal dot_key_q1     : std_logic;
   signal dot_key_q2     : std_logic;
   signal dot_key_q3     : std_logic;
   signal dot_confirm    : std_logic;
   signal dash_key_q1    : std_logic;
   signal dash_key_q2    : std_logic;
   signal dash_key_q3    : std_logic;
   signal dash_confirm   : std_logic;
   signal sidetone_out   : std_logic; 
   signal rx_strobe      : std_logic;
   signal tx_strobe      : std_logic;
   signal dot_sr         : std_logic_vector(4 downto 0);
   signal dash_sr        : std_logic_vector(4 downto 0);
   signal rx_dout        : unsigned(7 downto 0);
      
 

begin

  
   -- Instances of the components that make up the Keyer
   U100: uart_rx 
   port map (
      clk           =>  clk,  
      reset         =>  reset,
      rx            =>  uart_rx,
      strobe        =>  rx_strobe,
      dout          =>  rx_dout);
    
	U101: uart_tx
   port map (
      clk           =>  clk, 
      reset         =>  reset, 
      start_strobe  =>  tx_strobe,  
      wpm           =>  wpm,
      tx            =>  uart_tx);
   
   U102: sidetone
   port map (
      clk           =>  clk,
      reset         =>  reset,
      enable        =>  sidetone_out,
      spkr          =>  spkr); 
      
      
   
  
-------------------------------------------------------------
-- This sets the keyer speed based on receiving a
-- "<" or a ">" char over RS232
-------------------------------------------------------------
   process (reset, clk)
     
	begin
	   if (reset = '1') then
	      wpm <= 20; -- Default to 20 WPM
         tx_strobe <= '0';
                            
		elsif rising_edge(clk) then
         
         -- If not reset then set wpm to 5
         if (wpm = 0) then wpm <= 5;
         end if;
                  
         -- Define resting level for send strobe
         tx_strobe <= '0';
         
         if (rx_strobe = '1') then
            -- If RS232 char is a '>' then increase wpm
            -- If RS232 char is a '<' then idecrease wpm
            if (rx_dout = X"3E" and wpm /= 63) then
               wpm <= wpm + 1;
            elsif (rx_dout = X"3C" and wpm /= 5) then
               wpm <= wpm - 1;
            end if;
            -- Echo the speed over RS232 even if not changed
            tx_strobe <= '1';
            
         end if;
                         
      end if;
   end process;   


-------------------------------------------------------------
-- This is the state machine part that 
-- sets the state to the next state
-------------------------------------------------------------

   process (clk, reset)
   begin
      if (reset = '1') then
         state <= IDLE;
                           
      elsif rising_edge(clk) then
         state <= next_state;
      end if;         
          
   end process;
   
   
-------------------------------------------------------------
-- This is the state machine part that 
-- determines the next state
-------------------------------------------------------------
   process (clk, reset)
	   variable on_cnt : integer;
   begin
      if (reset = '1') then
         next_state <= IDLE;
         on_cnt := 0;
         dot_mem <= '0';
         dash_mem <= '0';
         key_out <= '0';
         sidetone_out <= '0';
         led0 <= '0';
         
      elsif rising_edge(clk) then
                
         case state is
            when IDLE =>
            -- In this state we are waiting for user to hit dot or dash
            key_out <= '0';
            sidetone_out <= '0';
            led0 <= '0';
            
            -- Convert WPM speed to a timer count value.  This is done
            -- here so it can change only when not sending a char
            dot_time <= wpm_lookup(wpm);
            dash_time <= 3 * dot_time;
            
            
            -- If paddle dot input or memory bit active then start sending dot
            if (dot_confirm = '1' or dot_mem = '1') then
               on_cnt := 0;
               dot_mem <= '0';
               next_state <= DOT;
            -- If paddle dash input or memory bit active then start sending dash
            elsif ((dash_confirm = '1') or (dash_mem = '1')) then
               on_cnt := 0;
               dash_mem <= '0';
               next_state <= DASH;
            end if;
      
            when DOT =>
            -- We are sending a dot
            key_out <= '1';
            sidetone_out <= '1';
            led0 <= '1';
            on_cnt := on_cnt + 1;
            -- Wait for one dot time based on wpm setting
            if (on_cnt >= dot_time) then
               on_cnt := 0;
               next_state <= SPACE_AFTER_DOT;   
            -- If user hits dash them remember it
            elsif (dash_confirm = '1') then
               dash_mem <= '1';
            end if;
								
            when DASH =>
            -- We are sending a dash            
            key_out <= '1';
            sidetone_out <= '1';
            led0 <= '1';
            on_cnt := on_cnt + 1;
            -- Wait for one dash time based on wpm setting
            if (on_cnt >= dash_time) then
               on_cnt := 0;
               next_state <= SPACE_AFTER_DASH;
            -- If user hits dot then remember it
            elsif (dot_confirm = '1') then
               dot_mem <= '1';
            end if;
               
            when SPACE_AFTER_DOT =>
            -- We are timing out the space after a dot 
            key_out <= '0';
            sidetone_out <= '0';
            led0 <= '0';
            on_cnt := on_cnt + 1;
            -- Wait for one dot time
            if (on_cnt >= dot_time) then
               on_cnt := 0;
               next_state <= IDLE;
            -- If uses hits dash then remember it
            elsif (dash_confirm = '1') then
               dash_mem <= '1';
            end if;
         
         			
            when SPACE_AFTER_DASH =>
            -- We are timing out the space after a dash 
            key_out <= '0';
            sidetone_out <= '0';
            led0 <= '0';
            on_cnt := on_cnt + 1;
            -- Wait for one dot time
            if (on_cnt >= dot_time) then
               on_cnt := 0;
               next_state <= IDLE;
            -- If uses hits dot then remember it
            elsif (dot_confirm = '1') then
               dot_mem <= '1';
            end if;
         
         
            when others =>
            -- Should not get here but if it does then clear everything
            key_out <= '0';
            sidetone_out <= '0';
            led0 <= '0';
            on_cnt := 0;
            dot_mem <= '0';
            dash_mem <= '0';
            next_state <= IDLE;			
      
         end case;
	   end if;	
   end process;


 
 
 

-------------------------------------------------------------------------------
-- This syncs and deglitches the paddle inputs, and inverts them.
-- This provides some glitch protection.  Do not need to fully debounce
-- paddle input because once a dot or dash is started, a few msec bounce
-- has no effect. 
--
-- This deglitch causes a delay between the paddle contact and the
-- keyer output of about 0.5 msec, not noticeable.  My MFJ 407D
-- keyer has a delay of 1 msec.
-------------------------------------------------------------------------------
   process (clk, reset)
      variable cnt : integer range 0 to 8191;
   
   begin
      if (reset = '1') then
	      dot_key_q1   <= '1'; -- Paddle inputs rest in high state 
         dot_key_q2   <= '1';
         dash_key_q1  <= '1';
         dash_key_q2  <= '1';
                  
         dot_sr <= "11111";
         dash_sr <= "11111";
                  
         dot_confirm  <= '0';
         dash_confirm <= '0';
         cnt := 0;
                                
		elsif rising_edge(clk) then
      
         cnt := cnt + 1;
         -- Divide down to a 10KHz rate so we sample paddle input every 
         -- 100 usec
         if (cnt = 5000) then
            cnt := 0;
            
            -- Register the asynch paddle inputs using the recommended
            -- two flip flops
            dot_key_q1 <= dot_key; 
            dash_key_q1 <= dash_key; 
            dot_key_q2 <= dot_key_q1;
            dash_key_q2 <= dash_key_q1;
            
            -- This uses a shift register to deglitch dot input
            -- Note this also inverts the paddle input
            -- Requires the input to be stable for 0.5 msec
            dot_sr <= dot_sr(3 downto 0) & dot_key_q2;
            if (dot_sr = "00000") then dot_confirm <= '1'; 
            elsif (dot_sr = "11111") then dot_confirm <= '0'; 
            end if; 
            
            -- This uses a shift register to deglitch dash input
            -- Note this also inverts the paddle input
            -- Requires the input to be stable for 0.5 msec
            dash_sr <= dash_sr(3 downto 0) & dash_key_q2;
            if (dash_sr = "00000") then dash_confirm <= '1'; 
            elsif (dash_sr = "11111") then dash_confirm <= '0'; 
            end if; 
            
                             
         end if;            
      end if;
   end process;
   
 

end rtl;


