-------------------------------------------------------------------------------
-- Keyer Project
-- Jim Brady   Jan 2011
-- The license is FreeBSD, see the file "copying.txt"
--
-- This module receives RS232 data.  Use 115.2K baud, 8 data bits
-- 1 stop bit, no parity

-- The line in bit idles at a '1' state, goes to '0' for the
-- start bit. A '1' data bit is sent as a high state (3V), and the
-- stop bit is a high state. It sends start bit, 8 data bits LSB
-- first, then stop bit
------------------------------------------------------------------------------


library IEEE;
   use IEEE.std_logic_1164.all;
   use IEEE.numeric_std.all;
	

entity uart_rx is
   port (
      clk        : in  std_logic;  -- 50MHz clock
      reset      : in  std_logic;  -- South button
      rx         : in  std_logic;  -- RS232 line in
      strobe     : out std_logic;  -- Char was received
      dout       : out unsigned(7 downto 0));  -- Received char

end uart_rx;



architecture rtl of uart_rx is

   signal rx_q1        : std_logic;
   signal rx_q2        : std_logic;
   signal rx_q3        : std_logic;
   signal local_dout   : unsigned(7 downto 0);
   signal local_strobe : std_logic;
    
begin

   
   
-------------------------------------------------------------------------------
-- This receives an RS232 character at 115K baud 
-- and issues a strobe pulse when a char has been received
-- The strobe is high for a single 50MHz clock
-------------------------------------------------------------------------------
   process (reset, clk)
      variable cnt     : integer range 0 to 511;
      variable bit_nr  : integer range 0 to 15;
      variable run_flg : boolean;
	   
	begin
	   if (reset = '1') then
	      cnt     := 0;
         bit_nr  := 0;
         local_strobe  <= '0';
         strobe <= '0';
         run_flg := FALSE;
         local_dout <= X"00";
     
		elsif rising_edge(clk) then
      
         -- Define the strobe resting level
         local_strobe <= '0';
		   
         -- Register the aynchronous input
         rx_q1 <= rx;
         rx_q2 <= rx_q1;
         rx_q3 <= rx_q2;
         
         -- Detect a high to low transition)
         if ((rx_q3 = '1') and (rx_q2 = '0') and (run_flg = FALSE)) then
            cnt := 217;
            bit_nr := 0;
            local_dout <= X"00";
            run_flg := TRUE;
            
         end if;   
         
         
                    
         if (run_flg) then
         	-- Divide 50MHz system clock down to 115.2K baud
            cnt := cnt + 1;
            if (cnt = 434) then
               cnt := 0;
               bit_nr := bit_nr + 1;
                        
               case bit_nr is
                  
                  -- Receive the 8 data bits, LSB first
                  when 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 =>
                  local_dout(bit_nr - 2) <= rx;
               
                  -- Done (this is middle of stop bit)
                  when 10 =>
                  run_flg := FALSE;
                  -- Strobe indicates that a char was received
                  local_strobe <= '1';
                         
                                               
                  when others =>
                  null;
               
               end case;  
            end if;
         end if;
         dout <= local_dout;
         strobe <= local_strobe;
      end if;
   end process;   


 
  
end rtl;

