-------------------------------------------------------------------------------
-- Keyer Project
-- Jim Brady   Jan 2011
-- The license is FreeBSD, see the file "copying.txt"
--
-- This module handles the RS232 transmit.  Use 115.2K baud, 8 data bits
-- 1 stop bit, no parity
--
-- The tx line out idles at a '1' state, then goes to '0' for the
-- start bit.  A '1' data bit is sent as a high state (3V), and the
-- stop bit is a high state. It sends start bit, 8 data bits LSB
-- first, then stop bit

-- To send, pulse the start line for high for 1 system clock.  Do not
-- change data while sending
-------------------------------------------------------------------------------


library IEEE;
   use IEEE.std_logic_1164.all;
   use IEEE.numeric_std.all;
  

entity uart_tx is
   port (clk           : in  std_logic;  -- 50MHz clock
         reset         : in  std_logic;  -- South button
         start_strobe  : in  std_logic;  -- Strobe to start sending
         wpm           : in  integer range 0 to 63;  -- speed in wpm 0 - 63
         tx            : out std_logic);  -- RS232 line data out
end uart_tx;

 
      
          
architecture rtl of uart_tx is


   -- This lookup table is used to convert a binary value to two
   -- decimal digits.  The high digit is in high nibble and low digit
   -- is in low nibble
   type bin2dec_type is array(0 to 63) of unsigned(7 downto 0);
   constant bin2dec_lookup: bin2dec_type :=
       (X"00", X"01", X"02", X"03", X"04", X"05", X"06", X"07", X"08", X"09",
		  X"10", X"11", X"12", X"13", X"14", X"15", X"16", X"17", X"18", X"19",
        X"20", X"21", X"22", X"23", X"24", X"25", X"26", X"27", X"28", X"29",
        X"30", X"31", X"32", X"33", X"34", X"35", X"36", X"37", X"38", X"39",
        X"40", X"41", X"42", X"43", X"44", X"45", X"46", X"47", X"48", X"49",
        X"50", X"51", X"52", X"53", X"54", X"55", X"56", X"57", X"58", X"59",
        X"60", X"61", X"62", X"63");
        
    
    
   signal char_to_send   : unsigned(7 downto 0);
   signal send_strobe    : std_logic;
   signal temp           : unsigned(7 downto 0);
   signal hi_ascii       : unsigned(7 downto 0);
   signal lo_ascii       : unsigned(7 downto 0);
   

  
begin

   -- Combinatorial code to convert speed in wpm into low digit and high digit
   -- Add 30 hex so the result is in ASCII
   temp <= bin2dec_lookup(wpm);
   hi_ascii <= "0011" & temp(7 downto 4);
   lo_ascii <= "0011" & temp(3 downto 0);
  

  
-------------------------------------------------------------------------------
-- This converts WPM to two ASCII chars then strobes the send UART
-- to send the chars. The strobe is high for a single 50MHz clock
-------------------------------------------------------------------------------
     process (reset, clk)
            
	   variable cnt       : integer range 0 to 32767;
      variable run_flg   : boolean;
      	   
   
	begin
	   if (reset = '1') then
	      cnt          := 0;
	      run_flg      := FALSE;
         send_strobe  <= '0';
                   
		elsif rising_edge(clk) then
		
         -- Define resting level
         send_strobe <= '0';
      
         if (start_strobe = '1') then
            cnt := 0;
            run_flg := TRUE;
         end if;
         
         if (run_flg) then
            cnt := cnt + 1;
                                    
            case cnt is
               when 2 =>
               -- Start on "2" instead of "1" to allow a little time
               -- Send first character
               char_to_send <= hi_ascii;
				   send_strobe <= '1';
            
               when 5000 =>              
               -- Wait 100 usec then send second character
               char_to_send <= lo_ascii;
               send_strobe <= '1';
               
               when 10000 =>              
               -- Wait 100 usec then send carriage return
               char_to_send <= X"0D";
               send_strobe <= '1';  
            
               when 15000 =>              
               -- Wait 100 usec then send line feed
               char_to_send <= X"0A";
               send_strobe <= '1'; 
               
               when 20000 =>  
               -- Done
               run_flg := FALSE; 
                            
               when others =>
               null;   
          
            end case;
         end if;
      end if;
   end process;



-------------------------------------------------------------------------------
-- This process sends a character over RS232 at 115K baud 
-- This is used as follows,
--   1. Put char in char_to_send, do not change
--   2. Strobe send strobe 
--   3. Wait for char to be sent
-------------------------------------------------------------------------------
   process (reset, clk)
      variable cnt     : integer range 0 to 511;
      variable bit_nr  : integer range 0 to 15;
      variable run_flg : boolean;
	   
	begin
	   if (reset = '1') then
	      cnt     := 0;
         bit_nr  := 0;
         run_flg := FALSE;
         tx      <= '1';
         
		elsif rising_edge(clk) then
		      
         -- Detect send strobe
         if (send_strobe = '1') then
            cnt := 0;
            bit_nr := 0;
            run_flg := TRUE;
         end if;
   
         if (run_flg) then
         	-- Divide 50MHz system clock down to 115.2K baud
            cnt := cnt + 1;
            if (cnt = 434) then
               cnt := 0;
               bit_nr := bit_nr + 1;
                        
               case bit_nr is
                  
                  -- Send start bit
                  when 1 =>
                  tx <= '0';

                  -- Send the 8 data bits, LSB first
                  when 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 =>
                  tx <= char_to_send(bit_nr - 2);
               
                  -- Send stop bit
                  when 10 =>
                  tx <= '1';
               
                  -- Stop sending
                  when 11 =>
                  run_flg := FALSE;
                                 
                  when others =>
                  null;
               
               end case;  
            end if;
         end if;   
      end if;
   end process;   

end rtl;

   
 